/*
 * Nombre del proyecto: Proyecto Arkanoid
 *
 * Nombre del fichero: lborrado.c
 *
 * Nombre de los autores: Grzegorz Adam Hankiewicz y Carlos Pereda Moreno
 *
 * Fecha: 18 de Enero del 2000
 *
 * Descripcin: Contiene las funciones de actualizacin de pantalla por
 * trozos. Se usan listas que almacenan qu partes de la pantalla cambian,
 * y cuales n. Luego slo se actualizan los cambios, optimizando la
 * velocidad del juego.
 */

/* CABECERAS GLOBALES */
#ifdef MSS
#include <mss.h>
#endif MSS
#include <stdlib.h>
#include <allegro.h>

/* CABECERAS DE FUNCIONES EXTERNAS */
#include "general.h"
#include "errores.h"
#include "struct.h"
#include "graph.h"

/* CABECERA PROPIA */
#define DECLARA_VARIABLES_LBORRADO_C
#include "lborrado.h"


/**[txh]********************************************************************

  Description: Funcin que crea las listas lgica y fsica que almacenan las
  coordenadas de los elementos en pantalla para su borrado y redibujado.

  Return: 0 si se crean las listas con xito; distinto de 0 en caso contrario.

***************************************************************************/


int      inicializa_listas_borrado (void)
{
   lista_borrado_logica = malloc (sizeof (LISTA_BORRADO));
   if (!lista_borrado_logica)
   {
      eprintf ("inicializa_lista_borrado: falta memoria #1");
      return 1;
   }
   lista_borrado_logica->next = NULL;

   lista_borrado_fisica = malloc (sizeof (LISTA_BORRADO));
   if (!lista_borrado_fisica)
   {
      eprintf ("inicializa_lista_borrado: falta memoria #2");
      return 1;
   }
   lista_borrado_fisica->next = NULL;

   return 0;
}
   

/**[txh]********************************************************************

  Description: Funcin que destruye las listas lgica y fsica de borrado
  y redibujado de los elementos.

***************************************************************************/

void     destruye_lista_borrado (void)
{
   if (lista_borrado_logica)
   {
      limpia_lista_borrado (lista_borrado_logica);
      free (lista_borrado_logica);
      lista_borrado_logica = NULL;
   }

   if (lista_borrado_fisica)
   {
      limpia_lista_borrado (lista_borrado_fisica);
      free (lista_borrado_fisica);
      lista_borrado_logica = NULL;
   }
}


/**[txh]********************************************************************

  Description: Funcin que vaca una lista lgica de elementos. Se llama
  cuando ya se ha procesado una lista y para la siguiente pasada se requiere
  tener la lista limpia.

***************************************************************************/


void     limpia_lista_borrado (LISTA_BORRADO *lista)
{
   LISTA_BORRADO *p,*p2;

   if (!lista)
   {
      eprintf ("limpia_lista_borrado: puntero nulo");
      return ;
   }

   p = lista->next;
   lista->next = NULL;

   while (p)
   {
      p2 = p->next;
      free (p);
      p = p2;
   }
}

/**[txh]********************************************************************

  Description: Esta funcin ser usada por las rutinas grficas para
  indicar que una zona de pantalla ha sido sobreescrita, y ser necesario
  redibujarla en el siguiente frame. Se usa por ejemplo desde las
  funciones que dibujan la pelota o la raqueta. Se le pasan las coordenadas
  lgicas de la pantalla virtual.

***************************************************************************/

void     marca_zona_borrado (LISTA_BORRADO *lista, int x, int y, int w, int h)
{
   LISTA_BORRADO *p;

   if (!lista)
   {
      eprintf ("marca_zona_borrado: puntero nulo");
      return ;
   }

   /* Reservamos memoria para la estructura */
   p = malloc (sizeof (LISTA_BORRADO));
   if (!p)
   {
      eprintf ("marca_zona_borrado: no hay memoria suficiente");
      return ;
   }

   /* Asignamos variables */
   p->x = x;
   p->y = y;
   p->w = w;
   p->h = h;

   /* Insertamos el nodo */
   p->next = lista->next;
   lista->next = p;

   /* comprobamos si hay que actualizar la pantalla */
   if (lista != lista_borrado_fisica)
      marca_zona_borrado (lista_borrado_fisica, x, y, w, h);
}


/**[txh]********************************************************************

  Description: Esta funcin recorre la lista de zonas marcadas para
  borrar, y por cada una de ellas llama a la de redibujado de ladrillos.
  Al acabar llama a @x{limpia_lista_borrado}, para el prximo borrado.

***************************************************************************/


void     recupera_zonas_borrado (DATOS_JUEGO *datos)
{
   LISTA_BORRADO *p = lista_borrado_logica->next;
   int x1, y1, x2, y2;

   while (p)
   {
      x1 = p->x / TAM_CELDA;
      y1 = p->y / TAM_CELDA;
      x2 = (p->w + p->x) / TAM_CELDA;
      y2 = (p->h + p->y)/ TAM_CELDA;
      x2++; y2++;

      /* Correcciones de rango */
      if (x1 < 0) x1 = 0;
      if (y1 < 0) y1 = 0;
      if (x2 > datos->ladrillos->w) x2 = datos->ladrillos->w;
      if (y2 > datos->ladrillos->h) y2 = datos->ladrillos->h;

      dibujar_ladrillos (datos, x1, y1, x2, y2);
      marca_zona_borrado (lista_borrado_fisica, x1*TAM_CELDA, y1*TAM_CELDA,
                          (x2-x1)*TAM_CELDA, (y2-y1)*TAM_CELDA);

      p = p->next;
   }

   limpia_lista_borrado (lista_borrado_logica);
}

/**[txh]********************************************************************

  Description: Esta funcin recorre la lista de zonas marcadas para
  dibujar, y por cada una de ellas llama a la de redibujado del bitmap bmp
  en pantalla.
  Al acabar llama a @x{limpia_lista_borrado}, para el prximo redibujado.

***************************************************************************/

void     recupera_zonas_pantalla (BITMAP *bmp, DATOS_JUEGO *datos)
{
   LISTA_BORRADO *p = lista_borrado_fisica->next;

   while (p)
   {
      /* Correcciones de rango */
      if (p->x < 0) p->x = 0;
      if (p->y < 0) p->y = 0;
      if (p->x >= datos->bmp->w) continue;
      if (p->y >= datos->bmp->h) continue;
      if (p->w > datos->bmp->w - p->x) p->w = datos->bmp->w - p->x;
      if (p->h > datos->bmp->h - p->y) p->h = datos->bmp->h - p->y;
      
      stretch_blit (datos->bmp, bmp, p->x, p->y, p->w, p->h,
                    margen_izq + p->x*escalado_pantalla,
                    margen_alto + p->y*escalado_pantalla,
                    p->w*escalado_pantalla, p->h*escalado_pantalla);
      p = p->next;
   }

   limpia_lista_borrado (lista_borrado_fisica);
}
