/*
 * Nombre del proyecto: Proyecto Arkanoid
 *
 * Nombre del fichero: log.c
 *
 * Nombre de los autores: Grzegorz Adam Hankiewicz y Carlos Pereda Moreno
 *
 * Fecha: 18 de Enero del 2000
 *
 * Descripcin: Funciones bsicas de logging. Permite mandar mensajes a un
 * fichero de texto y a una lista enlazada. Se podra aadir una funcin para
 * mostrar el contenido de la lista enlazada por pantalla, haciendo de
 * funcin 'consola de mensajes' grfica durante el juego (ya visto en juegos
 * como el Quake).
 */

/* CABECERAS GLOBALES */
#ifdef MSS
#include <mss.h>
#endif MSS
#include <stdio.h>
#include <string.h>
#include <allegro.h>
#include <time.h>

/* CABECERAS DE FUNCIONES EXTERNAS */
#include "general.h"
#include "errores.h"
#include "main.h"

/* CABECERA PROPIA */
#define DECLARA_VARIABLES_LOG_C
#include "log.h"


/**[txh]********************************************************************

  Description: Esta funcin prepara el mdulo de logging del programa. Dado
  que este juego no es una aplicacion crtica, y a que puede ser til para
  la depuracin/seguimiento del programa por cualquier persona, incluso
  ajena al proyecto, se usar un sistema de listas enlazadas que almacenarn
  una cola de mensajes variable, que puede ser impresa en pantalla, o
  almacenada en un fichero, etc...
  @*
  Si la inicializacion tiene xito, se hacen unos @x{lprintf}s con la fecha
  de comienzo de logging. Se pasa como parmetro el numero mximo de
  mensajes que se quiere tener en la lista enlazada. Si se pasa cero, se
  asumir un valor de uno. Se recomiendan valores de entre 50 y 200.

  Return: (int) Cero con xito, positivo en caso contrario

***************************************************************************/

int      inicializa_logging (unsigned long max_mensajes)
{
   time_t curtime;
   struct tm *loctime;
   
   /* Comprobamos si se llama ms de una vez... */
   if (GET_FLAG (banderas, BIT_LOGGING_ACTIVO))
   {
      eprintf ("inicializa_logging: el bit de log ya estaba activado.");
      return 3;
   }
   
   /* Probamos reservar el primer nodo 'tonto' */
   mensaje_primero = malloc (sizeof (MENSAJE));
   if (!mensaje_primero)
   {
      eprintf ("inicializa_logging #1: no hay suficiente memoria");
      return 1;
   }

   /* Limpiamos la memoria del nodo, y reservamos el segundo */
   memset (mensaje_primero, 0, sizeof (MENSAJE));

   mensaje_ultimo = malloc (sizeof (MENSAJE));
   if (!mensaje_ultimo)
   {
      eprintf ("inicializa_logging #2: no hay suficiente memoria");
      free (mensaje_primero);
      return 2;
   }

   /* Limpiamos el otro nodo, reasignamos los enlaces, y activamos flags */
   memset (mensaje_ultimo, 0, sizeof (MENSAJE));

   mensaje_primero->next = mensaje_ultimo;
   mensaje_ultimo->prev = mensaje_primero;

   /* Asignamos los valores de contador y mximo numero de mensajes */
   maximo_mensajes_log_permitidos = max_mensajes;
   if (!maximo_mensajes_log_permitidos)
      maximo_mensajes_log_permitidos++;
   num_mensajes_log = 0;
   
   SET_FLAG (banderas, BIT_LOGGING_ACTIVO);

   /* Vale, iniciamos el sistema con unos logs del tiempo de arranque */
   /* Conseguimos el tiempo actual. */
   curtime = time (NULL);
     
   /* Lo convertimos en tiempo local. */
   loctime = localtime (&curtime);
     
   /* Mandamos un mensaje con la fecha y cierta informacin extra */
   lprintf (" \nComenzando logging. Fecha actual: %s", asctime (loctime));
   imprime_informacion_sistema (1);
   return 0;
}


/**[txh]********************************************************************

  Description: Esta funcin libera la memoria usada por los mensajes de
  logging. Debe ser llamada antes de salir del juego, para garantizar la
  limpieza de memoria.

***************************************************************************/

void     finaliza_logging (void)
{
   time_t curtime;
   struct tm *loctime;
   
   /* Comprobamos que se haba iniciado el modulo de logs */
   if (!GET_FLAG (banderas, BIT_LOGGING_ACTIVO))
   {
      eprintf ("finaliza_logging: el bit de log estaba desactivado.");
      return;
   }

   /* Mandamos un mensaje con la hora */
   curtime = time (NULL);
   loctime = localtime (&curtime);
   lprintf ("finaliza_logging: Cerrando. Fecha %s", asctime (loctime));

   /* Destruimos primero las cadenas del medio */
   libera_mensajes_log();

   /* Ahora destruimos los terminadores */
   destruye_mensaje_log (mensaje_primero);
   destruye_mensaje_log (mensaje_ultimo);

   mensaje_primero = mensaje_ultimo = NULL;

   /* Limpiamos banderas, y guardamos configuracin en caso necesario */
   CLEAR_FLAG (banderas, BIT_LOGGING_ACTIVO);
   if (GET_FLAG (banderas, BIT_LOG_A_FICHERO))
      set_config_string (SECCION_ARKANOID, CFG_FICH_LOG, path_fichero_log);
}

/**[txh]********************************************************************

  Description: Esta funcin libera la memoria usada por los mensajes que
  estan en la lista enlazada, sin destruir los terminadores.

***************************************************************************/

void     libera_mensajes_log (void)
{
   MENSAJE *p1,*p2;

   p1 = mensaje_primero->next;
   while (p1->texto)
   {
      p2=p1->next;
      destruye_mensaje_log (p1);
      p1=p2;
   }
}


/**[txh]********************************************************************

  Description: Esta funcin destruye un mensaje de la cadena, reenlazndola
  despus si es posible con sus vecinos.

***************************************************************************/


void     destruye_mensaje_log (MENSAJE *men)
{
   MENSAJE *prev,*next;

   if (!men)
   {
      /* Situacin anmala */
      eprintf ("destruye_mensaje_log: error, puntero nulo");
      return;
   }

   prev = men->prev;
   next = men->next;

   /* Destruimos la cadena de texto del mensaje, luego el enlace de la cadena.*/
   if (men->texto)
      free (men->texto);

   free (men);

   /* En caso de que existiesen 'vecinos', los reenlazamos.*/
   if (prev)
      prev->next = next;
   if (next)
      next->prev = prev;

   num_mensajes_log--;
}



/**[txh]********************************************************************

  Description: funcin para hacer logging. Aade un mensaje ms a la cadena
  de mensajes por el final. Si esta especificado globalmente, aade el
  mensaje automticamente a un fichero de texto especificado por la variable
  path_fichero_log. El mensaje de logging no se mostrar a no ser que el
  programa est mostrando los diversos mensajes en alguna zona de la
  pantalla, manualmente. Adems, si el bit de logging no esta activado, la
  peticin sera descartada sin mostrar error alguno.
  @*
  Si se usa \n a mitad de una cadena, sta ser troceada y en realidad se
  crearn dos mensajes o ms. Esta funcin es en realidad un "filtro"
  para @x{crea_mensaje}, que trata la cadena y la trocea si es necesario. Para
  crear una lnea vaca, no basta con \n, hay que poner al menos un espacio
  como carcter. Tras aadir el mensaje, llama a
  @x{elimina_excedente_lista_log}. No hace falta acabar los mensajes con \n,
  porque el ltimo \n es automatico.

  Example:
  
  @x{lprintf} ("Partida iniciada.\nMemoria restante libre %d", mem_bytes_libres);

***************************************************************************/

void     lprintf (char *msg, ...)
{
   char buf[STR],*token;
   va_list ap;

   /* Comprobamos si el logging esta activo, o no. Pero no damos error o
    * algo parecido, por si el logging se queria desactivar manualmente
    * durante un periodo de tiempo manual, modificando el bit global, o
    * deliberadamente se ignor iniciar el modulo de logging.
    */
   if (!GET_FLAG (banderas, BIT_LOGGING_ACTIVO))
      return;

   /* Transformacin de parmetros en cadena simple */
   va_start (ap, msg);
   uvsprintf (buf, msg, ap);
   va_end (ap);

   /* Troceamos la cadena y vemos si hay que hacer algo... */
   token = ustrtok (buf, "\n");
   crea_mensaje_log (token);
   
   while ((token = ustrtok (NULL, "\n")))
      crea_mensaje_log (token);

   elimina_excedente_lista_log();
}



/**[txh]********************************************************************

  Description: Esta es la rutina que hace el trabajo sucio de @x{lprintf}.
  Se le pasan cadenas de texto sin caracteres \n, y crea en memoria el objeto
  de la lista enlazada con el texto, y aade el mensaje a un fichero de
  logging en caso de que esa opcin estuviese activada. Es una funcin
  interna, no usar directamente.

***************************************************************************/

void     crea_mensaje_log (char *texto)
{
   MENSAJE *men;
   FILE *fichero;

   if (!texto)
   {
      eprintf ("crea_mensaje_log: puntero nulo");
      return;
   }

   /* Comprobamos si hay que salvar la cadena en un fichero de log */
   if (GET_FLAG (banderas, BIT_LOG_A_FICHERO))
   {
      fichero = fopen (uconvert_toascii (path_fichero_log, NULL), "at");
      if (!fichero)
      {
         eprintf ("crea_mensaje_log: no se pudo abrir el fichero de log %s",
                  path_fichero_log);
         return;
      }
      fprintf (fichero, "%s\n", texto);
      fclose (fichero);
   }

   /* Reservamos memoria para el objeto de la cadena */
   men = malloc (sizeof (MENSAJE));
   if (!men)
   {
      eprintf ("crea_mensaje_log: #1 no hay suficiente memoria libre. Se queria "
               "mostrar:\n%s", texto);
      return;
   }

   /* Ahora con la cadena de texto */
   men->texto = malloc (strlen (texto)+1);
   if (!men->texto)
   {
      free (men);
      eprintf ("crea_mensaje_log: #2 no hay suficiente memoria libre. Se queria "
               "mostrar:\n%s", texto);
   }
   ustrcpy (men->texto, texto);

   /* Insertamos el nodo al final. */
   men->next = mensaje_ultimo;
   men->prev = mensaje_ultimo->prev;

   /* Reenlazamos sus hermanos e incrementamos el contador */
   men->next->prev = men;
   men->prev->next = men;
   num_mensajes_log++;
}


/**[txh]********************************************************************

  Description: Comprueba si nombre_fichero es un fichero de log que se
  puede usar, y ajusta la variable path_fichero_log en consecuencia. Esta
  rutina llama siempre a eprintf si hay algun inconveniente, como que ya
  hay un fichero log siendo usado (solo puede haber uno). Si no hay
  problemas, se activa el BIT_LOG_A_FICHERO. Se presupone que el texto del
  fichero _YA_ se pasa como UTF-8, y NO ASCII. Tener cuidado con eso.
  Si el nombre del fichero de logging es NULL, se desactivan los logs.

  Return: (int) Cero con xito, un numero positivo con error.

***************************************************************************/

int      comprueba_validez_fichero_log (char *nombre_fichero)
{
   FILE *fichero;

   if (!nombre_fichero)
   {
      eprintf ("comprueba_validez_fichero_log: puntero nulo");
      return 1;
   }

   /* Comprobamos si se quiere anular el fichero de logging */
   if (ustricmp (nombre_fichero, "NULL") == 0)
   {
      SET_FLAG (banderas, BIT_IGNORA_LOG);
      CLEAR_FLAG (banderas, BIT_LOG_A_FICHERO);
      return 0;
   }

   /* Estamos repitiendo esto? */
   if (GET_FLAG (banderas, BIT_LOG_A_FICHERO))
   {
      eprintf ("comprueba_validez_fichero_log: no se puede usar %s",
               nombre_fichero);
      eprintf ("comprueba_validez_fichero_log: %s ya esta siendo usado",
               path_fichero_log);
      return 2;
   }

   /* Validamos el fichero abrindolo si no es NULL */
   fix_filename_path (path_fichero_log, nombre_fichero, STR);
   fichero = fopen (uconvert_toascii (path_fichero_log, NULL), "at");
   if (!fichero)
   {
      eprintf ("comprueba_validez_fichero_log: No se puede usar %s como "
               "fichero de logging.", path_fichero_log);
      return 3;
   }
   fclose (fichero);
   
   /* Todo verificado, usamos el fichero de log */
   SET_FLAG (banderas, BIT_LOG_A_FICHERO);
   return 0;
}



/**[txh]********************************************************************

  Description: Esta funcin va eliminando mensajes de la lista enlazada,
  siempre y cuando el nmero total sea mayor o igual al numero mximo de
  mensajes permitidos en la lista. Funcin interna.

***************************************************************************/

void     elimina_excedente_lista_log (void)
{
   while (num_mensajes_log >= maximo_mensajes_log_permitidos)
      destruye_mensaje_log (mensaje_primero->next);
}



/**[txh]********************************************************************

  Description: Funcin genrica que muestra informacin sobre el sistema
  gracias a las variables de Allegro. Puede ser llamada para mostrar la
  informacin en una pantalla de texto (antes o tras un modo grfico), o
  para mandar la informacin al sistema de logging. Para hacer esto ltimo,
  el parmetro de la funcin debe ser distinto de cero.

  Example:
  @<pre>
  int main (void)
  {
    allegro_init();
    imprime_informacion_sistema (0);
    return 0;
  }
  @</pre>

***************************************************************************/

void     imprime_informacion_sistema (unsigned char salida_lprintf)
{
   void  (*mensaje)(char *msg, ...);

   if (salida_lprintf)
      mensaje = lprintf;
   else
      mensaje = allegro_message;
   
   mensaje (NOMBRE_PROYECTO ", version %d.%d: " AUTORES ".\n",
            ARKANOID_VERSION, ARKANOID_SUB_VERSION);

   mensaje ("Compilado con %s.\nMaquina %s\n", allegro_id, cpu_vendor);
   switch (cpu_family)
   {
      case 3: mensaje ("CPU: 386\n"); break;
      case 4: mensaje ("CPU: 486\n"); break;
      case 5: mensaje ("CPU: 586 o Pentium\n"); break;
      case 6: mensaje ("CPU: 686, PentiumPro o superior\n"); break;
      default:mensaje ("CPU desconocida\n"); break;
   }

   switch (os_type)
   {
      case OSTYPE_UNKNOWN: mensaje ("SO: desconocido o msdos puro\n"); break;
      case OSTYPE_WIN3:    mensaje ("SO: Windows 3.1 o previo\n"); break;
      case OSTYPE_WIN95:   mensaje ("SO: Windows 95\n"); break;
      case OSTYPE_WIN98:   mensaje ("SO: Windows 98\n"); break;
      case OSTYPE_WINNT:   mensaje ("SO: Windows NT (argh!)\n"); break;
      case OSTYPE_OS2:     mensaje ("SO: OS/2\n"); break;
      case OSTYPE_WARP:    mensaje ("SO: OS/2 Warp\n"); break;
      case OSTYPE_DOSEMU:  mensaje ("SO: Linux DOSEMU\n"); break;
      case OSTYPE_OPENDOS: mensaje ("SO: Caldera OpenDos\n"); break;
      case OSTYPE_LINUX:   mensaje ("SO: Linux\n"); break;
      case OSTYPE_UNIX:    mensaje ("SO: Variante Unix desconocida\n"); break;
   }
   mensaje ("-----\n");
}


