/*
 * Nombre del proyecto: Proyecto Arkanoid
 *
 * Nombre del fichero: logico.c
 *
 * Nombre de los autores: Grzegorz Adam Hankiewicz y Carlos Pereda Moreno
 *
 * Fecha: 18 de Enero del 2000
 *
 * Descripcin: Contiene gran parte de las rutinas a nivel lgico del juego,
 * como son las colisiones, los casos de muerte, o la entrada de usuario
 * mediante teclado.
 */

/* CABECERAS GLOBALES */
#ifdef MSS
#include <mss.h>
#endif MSS
#include <string.h>
#include <allegro.h>

/* CABECERAS DE FUNCIONES EXTERNAS */
#include "general.h"
#include "struct.h"
#include "errores.h"
#include "log.h"
#include "main.h"
#include "lborrado.h"

/* CABECERA PROPIA */
#define DECLARA_VARIABLES_LOGICO_C
#include "logico.h"



/**[txh]********************************************************************

  Description: Esta funcin es genrica para detectar las colisiones tanto
  de los bordes de la pantalla como de la raqueta como de los ladrillos.
  Acepta como parmetros la posicin de pantalla que hay que comprobar.

  Return: (int) Cero si no hubo colisin. Cualquier otro nmero con colisin.

***************************************************************************/

int      hay_colision (DATOS_JUEGO *datos, fixed dx, fixed dy)
{
   int xl,yl,valor;
   fixed x,y;

   x = datos->x_p + dx;
   y = datos->y_p + dy;

   /* Chequeos con los bordes de la pantalla */
   if (x >= itofix (RES_LOGICA_X))
      return 1;

   if (x < 0)
      return 2;

   if (y < 0)
      return 3;

   /* Chequeos con la raqueta */
   if ((y >= F_ALTURA_RAQUETA && (y <= F_ALTURA_RAQUETA+itofix (ALTURA_RAQUETA)))
       && (x <= datos->x_r + F_ANCHO_RAQUETA/2)
       && (x >= datos->x_r - F_ANCHO_RAQUETA/2))
   {
      SET_FLAG (banderas, BIT_COLISION_RAQUETA);
      return 4;
   }

   /* Chequeos con los ladrillos */
   xl = (x / TAM_CELDA) >> 16;
   yl = (y / TAM_CELDA) >> 16;

   if (xl < 0 || yl < 0 || xl >= datos->ladrillos->w || yl >= datos->ladrillos->h)
      return 0;
   
   valor = getpixel (datos->ladrillos, xl, yl);
   if (valor != 0)
      return 6;

   return 0;
}

/**[txh]********************************************************************

  Description: Detecta si la pelota colisiona con los laterales de la pantalla,
  la parte superior de sta, o con la raqueta, en cuyo caso cambiar de
  direccin evitando salirse de la pantalla. Aparte de modificar los valores
  de dx y dy, tambin hay que modificar el ngulo de datos, para que desde
  fuera el cambio de direccin afecte al bucle de desplazamiento.

***************************************************************************/

int      detectar_colision_pelota (fixed *dx, fixed *dy, DATOS_JUEGO *datos, int volver)
{
   fixed angulo, ax, ay;
   int col_x, col_y;

   /* Bucle que recorre la circunferencia */
   for (angulo = 0; angulo <= itofix (256); angulo += INC_ANG_COLISION)
   {
      ax = RAD_BOLA * fcos (angulo);
      ay = RAD_BOLA * fsin (angulo) * -1;

      if (hay_colision (datos, ax + *dx, ay + *dy))
      {
         if (volver)
            return 1;

         col_x = (datos->x_p + ax + *dx) >> 16;
         col_y = (datos->y_p + ay + *dy) >> 16;

         lprintf ("colision #1 en %dx%d, angulo_col: %d, angulo %d, direccion "
                  "%d %d", col_x, col_y, fixtoi (angulo),
                  fixtoi (datos->angulo), *dx, *dy);
         /* Determinamos el futuro movimiento comprobando si movisemos bien... */
         /* Negamos el eje x y vemos si hay colisin */
         *dx *= -1;

         if (detectar_colision_pelota (dx, dy, datos, volver+1))
         {
            /* Si que hay colisin, por lo que recuperamos datos y negamos eje y */
            *dx *= -1;
            *dy *= -1;
            datos->angulo = itofix (256) - datos->angulo;
            lprintf ("colision #2 en %dx%d, movimiento vertical permitido: "
                     "dx:%d dy:%d, nuevo angulo %d", fixtoi (*dx),
                     fixtoi(*dy), fixtoi(datos->angulo));
            datos->longitud += INC_VELOCIDAD;
            elimina_ladrillo (datos, col_x / TAM_CELDA, col_y / TAM_CELDA, -1);
            return 0;
         }

         /* No hubo problemas, por lo que seguimos hacia adelante */
         datos->angulo = itofix (128) - datos->angulo;
         lprintf ("colision #3 en %dx%d, movimiento horizontal permitido: "
                  "dx:%d dy:%d, nuevo angulo %d", fixtoi (*dx), fixtoi(*dy),
                  fixtoi(datos->angulo));
         elimina_ladrillo (datos, col_x / TAM_CELDA, col_y / TAM_CELDA, -1);
         datos->longitud += INC_VELOCIDAD;
         return 0;
      }
   }

   return 0;
}


/**[txh]********************************************************************

  Description: Funcin compleja que tiene como tarea mover la pelota en
  la direccin indicada por el ngulo y la longitud. Lo que hace la funcin
  es 'trocear' la longitud en pequeos incrementos de un pixel e ir
  comprobando las colisiones en cada segmento. De esta forma nos evitamos
  saltar paredes y las colisiones sern visualmente perfectas, aunque siempre
  habr errores despreciables en los clculos.
  @*
  La funcin va moviendo la pelota, y por cada segmento llama a la funcin
  @x{detectar_colision_pelota}, que determina las colisiones usando los
  incrementos dx y dy.

***************************************************************************/

#define MASCARA         0x00ffffff

void     avanza_pelota (DATOS_JUEGO *datos)
{
   fixed dx,dy,aux_long = datos->longitud;
   int old_x, old_y;

   old_x = old_y = -1;

   /* Copiada la longitud, procedemos a decrementarla en el bucle */
   while (aux_long > 0)
   {
      /* Es la longitud todava mayor que uno? */
      if (aux_long > itofix (1))
      {
         dx = fcos (datos->angulo);
         dy = fsin (datos->angulo) * -1;
      }
      /* O es menor que uno? */
      else
      {
         dx = fmul (aux_long, fcos (datos->angulo));
         dy = fmul (aux_long, fsin (datos->angulo)) * -1;
      }

      /* Comprobacin para no detectar colisiones en el mismo pixel */
      if (fixtoi (datos->x_p + dx) != old_x || fixtoi (datos->y_p + dy) != old_y)
      {
         /* Detectamos las posibles colisiones */
         detectar_colision_pelota (&dx, &dy, datos, 0);

         old_x = fixtoi (datos->x_p + dx);
         old_y = fixtoi (datos->y_p + dy);

         if (GET_FLAG (banderas, BIT_COLISION_RAQUETA))
         {
            datos->angulo = datos->angulo & MASCARA;
            
            if (datos->mov_x != 0 && datos->angulo < itofix(128))
            {
               if (datos->mov_x > 0)
               {
                  /* La raqueta se mueve a la derecha */
                  datos->angulo += itofix(8);

                  if (datos->angulo > itofix(64-8) && datos->angulo < itofix(64+8))
                     datos->angulo = itofix(64-8);

                  if (datos->angulo > itofix(128-8))
                     datos->angulo = itofix(128-8);
               }
               else
               {
                  /* La raqueta se mueve a la izquierda */
                  datos->angulo -= itofix(8);

                  if (datos->angulo < itofix(8))
                     datos->angulo -= itofix(8);

                  if (datos->angulo > itofix(64-8) && datos->angulo < itofix(64+8))
                     datos->angulo = itofix(64+8);
               }
            }
            CLEAR_FLAG (banderas, BIT_COLISION_RAQUETA);
         }
      }

      /* Desplazamos la pelota en su (nueva) direccin */
      datos->x_p += dx;
      datos->y_p += dy;

      /* Decrementamos la copia que almacena la longitud */
      aux_long -= itofix (1);
   }
}

/**[txh]********************************************************************

  Description: Detecta si la pelota se "escapa" por el lmite inferior de
  la pantalla, en cuyo caso activar la variable parar_juego.

***************************************************************************/

void     detectar_muerte_pelota (DATOS_JUEGO *datos)
{
   if (datos->y_p - F_RAD_BOLA >= itofix (RES_LOGICA_Y))
      SET_FLAG (datos->parar_juego, BIT_PERDER_VIDA);
}

/**[txh]********************************************************************

  Description: Detecta si la raqueta a llegado a los lmites izquierdo o
  derecho de la pantalla. => Si la raqueta llega a los extremos izquierdo o
  derecho de la pantalla el incremento de la raqueta no variar.

***************************************************************************/

void     detectar_colision_raqueta (DATOS_JUEGO *datos)
{
   if (datos->x_r + datos->dr <= F_ANCHO_RAQUETA/2)
      datos->dr = 0;
   if (datos->x_r + datos->dr >= itofix (datos->bmp->w-ANCHO_RAQUETA/2))
      datos->dr = 0;
}

/**[txh]********************************************************************

  Description: Mueve la raqueta incrementando/decrementando x_r con dr
  segn sea esta positiva o negativa.

***************************************************************************/

void     mover_raqueta (DATOS_JUEGO *datos)
{
   fixed desplazamiento;

   desplazamiento = datos->mov_x * (itofix (2) / MAX_MOV);

   if (!colision_pelota_raqueta (datos, desplazamiento))
      datos->x_r += desplazamiento;

   /* Colisiona la raqueta con la pared izquierda? */
   if (datos->x_r <= F_ANCHO_RAQUETA/2) datos->x_r = F_ANCHO_RAQUETA/2;
   /* O con la pared derecha? */
   if (datos->x_r >= itofix (datos->bmp->w-ANCHO_RAQUETA/2))
      datos->x_r = itofix (datos->bmp->w-ANCHO_RAQUETA/2);
}

/**[txh]********************************************************************

  Description: Funcin que recibe *datos y llama a todas las funciones de
  deteccin de colisiones, y a las de mover la raqueta y la pelota.

***************************************************************************/

void     bucle_logico (DATOS_JUEGO *datos)
{
   avanza_pelota (datos);
   detectar_muerte_pelota (datos);
   mover_raqueta (datos);
}

/**[txh]********************************************************************

  Description: Detecta si el usuario pulsa las teclas que mueven la raqueta,
  en cuyo caso incrementar/decrementar la variable dr, que mueve la raqueta
  a izquierda o a derecha segn sea esta positiva o negativa. Si el usuario
  pulsa ESC, incrementar la variable parar_juego.

***************************************************************************/



void     recoge_entrada_usuario (DATOS_JUEGO *datos)
{
   if (!datos)
   {
      eprintf ("recoge_entrada_usuario: puntero nulo");
      return ;
   }

   datos->mov_x = datos->mov_y = datos->botones = 0;

   poll_keyboard();
   
   /* Desplazamiento hacia la izquierda */
   if (key[KEY_LEFT])
      datos->mov_x = -MAX_MOV;
      
   /* Desplazamiento hacia la derecha */
   if (key[KEY_RIGHT])
      datos->mov_x = MAX_MOV;
      
   /* Desplazamiento hacia arriba */
   if (key[KEY_UP])
      datos->mov_y -= MAX_MOV;
      
   /* Desplazamiento hacia la abajo */
   if (key[KEY_DOWN])
      datos->mov_y += MAX_MOV;
      
   /* Chequeo de pulsacin de la tecla [ENTER] */
   if (key[KEY_ENTER])
      SET_FLAG (datos->botones, BT_DISP1);
      
   /* Si se pulsa F10 activaremos el bit de acceso al siguiente nivel */
   if (key[KEY_F10])
   {
      SET_FLAG (datos->parar_juego, BIT_SIGUIENTE_NIVEL);
      clear_keybuf();
   }

   /*Chequeo de pulsacin de la tecla [ESC] */
   if (key[KEY_ESC])
   {
      SET_FLAG (datos->parar_juego, BIT_SALIR_DE_JUEGO);
      lprintf ("recoge_entrada_usuario: el usuario pulso la tecla de escape");
   }
}




/**[txh]********************************************************************

  Description: Esta funcin elimina un ladrillo del mapa tras una colisin.
  Se le pasa la posicin x, y en coordenadas lgicas ya divididas por
  TAM_CELDA, y entonces se comprobar realmente si haba un ladrillo.
  Se usa recursividad para eliminar los ladrillos adyacentes simulando una
  funcin de rellenado grfico con el color cero.

***************************************************************************/

void     elimina_ladrillo (DATOS_JUEGO *datos, int col_x, int col_y, int col)
{
   if (!datos)
   {
      eprintf ("elimina_ladrillo: puntero nulo");
      return ;
   }

   if (!datos->ladrillos)
   {
      eprintf ("elimina_ladrillo: el mapa de ladrillos no existe (!)");
      return ;
   }

   if (col_x < 0 || col_y < 0 || col_x >= datos->ladrillos->w ||
       col_y >= datos->ladrillos->h)
      return ;

   /*  Empezamos la recursin ? */
   if (col < 0)
   {
      /* Caso base */
      col = getpixel (datos->ladrillos, col_x, col_y);

      if (col != 0)
      {
         elimina_ladrillo (datos, col_x, col_y, col);
         comprueba_si_quedan_ladrillos (datos);
      }
   }
   else
   {
      /* Caso recursivo: chequeo pixels adyacentes */
      if (getpixel (datos->ladrillos, col_x, col_y) == col)
      {
         lprintf ("elimina_ladrillo: borrando %d,%d de color %d", col_x,
                  col_y, col);
         putpixel (datos->ladrillos, col_x, col_y, 0);
         marca_zona_borrado (lista_borrado_logica, col_x*TAM_CELDA, col_y*TAM_CELDA, 1, 1);
         datos->puntos ++;
         elimina_ladrillo (datos, col_x+0, col_y+1, col);
         elimina_ladrillo (datos, col_x+1, col_y+0, col);
         elimina_ladrillo (datos, col_x-1, col_y+0, col);
         elimina_ladrillo (datos, col_x+0, col_y-1, col);
      }
   }
}


/**[txh]********************************************************************

  Description: Esta funcin recorre con un bucle doble anidado el mapa de
  ladrillos e incrementa la variable parar_juego si es necesario, indicando
  adems mediante el bit BIT_SIGUIENTE_NIVEL si se acab el nivel con xito.

***************************************************************************/

void     comprueba_si_quedan_ladrillos (DATOS_JUEGO *datos)
{
   int x, y, contador = 0;

   if (!datos)
   {
      eprintf ("comprueba_si_quedan_ladrillos: puntero nulo");
      return ;
   }

   if (!datos->ladrillos)
   {
      eprintf ("comprueba_si_quedan_ladrillos: el mapa de ladrillos no existe (!)");
      return ;
   }

   for (y = 0; y < datos->ladrillos->h; y++)
      for (x = 0; x < datos->ladrillos->w; x++)
         contador += datos->ladrillos->line[y][x];

   if (!contador)
   {
      SET_FLAG (datos->parar_juego, BIT_SIGUIENTE_NIVEL);
   }
}



/**[txh]********************************************************************

  Description: Detecta si habr colisin horizontal con la pelota. Esto se
  hace simplemente mirando si el rectngulo de la raqueta sobreescribe al
  de la pelota.

  Return: (int) cero si no hubo colisin.

***************************************************************************/

int      colision_pelota_raqueta (DATOS_JUEGO *datos, fixed desplazamiento)
{
   if (!datos) 
   {
      eprintf ("colision_pelota_raqueta: puntero nulo");
      return 1;
   }

   /* Chequeo rpido para descartar colisiones */
   if (datos->y_p + F_RAD_BOLA < F_ALTURA_RAQUETA)
      return 0;

   /* Chequeo margen izquierda */
   if (datos->x_p + F_RAD_BOLA < datos->x_r - F_ANCHO_RAQUETA/2 + desplazamiento)
      return 0;

   /* Chequeo margen derecha */
   if (datos->x_p - F_RAD_BOLA > datos->x_r + F_ANCHO_RAQUETA/2 + desplazamiento)
      return 0;

   return 1;
}


