/*
 * Nombre del proyecto: Proyecto Arkanoid
 *
 * Nombre del fichero: init.c
 *
 * Nombre de los autores: Grzegorz Adam Hankiewicz y Carlos Pereda Moreno
 *
 * Fecha: 18 de Enero del 2000
 *
 * Descripcin: Contiene las funciones que inicializan el sistema de erorres,
 * mensajes, modo grfico, etc. Tambin las funciones que cierran el sistema.
 */

/* CABECERAS GLOBALES */
#ifdef MSS
#include <mss.h>
#endif MSS
#include <stdio.h>
#include <allegro.h>

/* CABECERAS DE FUNCIONES EXTERNAS */
#include "general.h"
#include "struct.h"
#include "errores.h"
#include "main.h"
#include "log.h"
#include "juego.h"
#include "graph.h"
#include "hiscore.h"

/* CABECERA PROPIA */
#define DECLARA_VARIABLES_INIT_C
#include "init.h"


/**[txh]********************************************************************

  Description: Incrementa el contador de ticks. Esta funcin se ejecuta como
  una interrupcin a la velocidad que nosotros indiquemos mediante
  install_int_ex. La funcin debe hacer cuanto menos mejor para no "chupar"
  recursos a la CPU.

***************************************************************************/


void     incrementar_ticks (void)
{
   tick_counter++;
}
END_OF_FUNCTION (incrementar_ticks);

/**[txh]********************************************************************

  Description: Incrementa el contador de segundos. Esta funcin se ejecuta
  como una interrupcin a la velocidad que nosotros indiquemos mediante
  install_int_ex. La funcin debe hacer cuanto menos mejor para no "chupar"
  recursos a la CPU.

***************************************************************************/

void     incrementar_secs (void)
{
   sec_counter++;
}
END_OF_FUNCTION (incrementar_secs);


/**[txh]********************************************************************

  Description: Esta funcin se encarga de poner en orden todo lo relacionado
  con el juego. Primero inicializa algunas funciones de allegro, teclado e
  interrupciones. A continuacin lee la linea de comando en busca de
  parmetros tiles, marcando diferentes bits de estados en la bandera global.
  Luego, intenta entrar en un modo grfico, desde el cual ya es posible
  mostrar mensajes de una forma portable en cualquier plataforma. Excepto la
  decodificacin de la linea de comando, y el acto de entrar en modo grfico,
  el resto del programa puede usar el modo grfico para mostrar mensajes, ya
  sea de forma propia o mediante @x{lprintf}.

  Return: (int) Devuelve cero con xito, cualquier otro nmero si la rutina
  considera que el programa no debera continuar su ejecucin, por error
  grave.
  
***************************************************************************/

int      inicializa_programa (int argc,char *argv[])
{
   int ret;

   #ifdef MSS
   MSS_DISABLE_LOG_OUTPUT;
   #endif MSS
   /* Inicializacin de Allegro, teclado e interrupciones */
   allegro_init();
   install_keyboard();
   poll_keyboard(); /* Para entrar en modo polling automticamente */
   install_timer();
   install_mouse(); /* Para que haya cursor en la seleccin de resolucin */

   LOCK_VARIABLE (tick_counter);     /* Ajustamos variables y velocidad de la */
   LOCK_FUNCTION (incrementar_ticks);            /* funcin incrementar_ticks */
   LOCK_VARIABLE (sec_counter);      /* Ajustamos variables y velocidad de la */
   LOCK_FUNCTION (incrementar_secs);              /* funcin incrementar_secs */
   install_int_ex (incrementar_ticks, BPS_TO_TIMER (100));/* a 100 ticks/seg */
   install_int_ex (incrementar_secs, SECS_TO_TIMER (1));  /* a 1 tick/seg */

   /* Inicializamos el sistema de errores eprintf () */
   inicializa_mensajes_de_error();

   /* Lee la lnea de comandos, ajustando flags y variables */
   ret = decodifica_linea_de_comandos (argc, argv);
   if (ret)
      return ret;

   /* Ahora lee los valores por defecto del fichero .cfg */
   lee_configuracion_del_fichero_cfg();

   /* No entramos en modo grfico si se pide ayuda */
   if (!GET_FLAG (banderas, BIT_SHOW_HELP))
   {
      ret = entra_en_modo_grafico();
      if (ret)
         return ret;
   }

   /* Inicializamos el sistema de mensajes lprintf() */
   ret = inicializa_logging (NUM_MENSAJES_LOG);
   if (ret)
      return ret;

   /* Inicializa la tabla de rcords */
   init_hiscore();

   return 0;
}


/**[txh]********************************************************************

  Description: Esta funcin finaliza todo el juego, liberando memoria,
  escribiendo ficheros (si es necesario), y devolviendo al ordenador a un
  modo aceptable para el usuario (es decir, usable). Es la funcin que hay
  que llamar antes de salir, para asegurarnos de que no queda ningn cabo
  suelto. Si se necesita hacer una salida de emergencia con exit (num_error),
  es recomendable llamar antes a esta funcin.

***************************************************************************/

void     finaliza_programa (void)
{
   finaliza_logging();
   sal_del_modo_grafico();

   /* Desactivamos hardware controlado por Allegro */
   remove_mouse();
   remove_timer();
   remove_keyboard();

   /* Salvamos los rcords */
   shutdown_hiscore();

   /* Ahora que ya estaramos en modo texto, mostramos errores (si los hay) */
   muestra_mensajes_de_error();
   muestra_ayuda_linea_de_comando();
   allegro_exit();

   #ifdef MSS
   MSS_LOG_BLOCK_LIST;
   #endif
}


/**[txh]********************************************************************

  Description: Recorre la linea de argumentos intentando entenderlos,
  activando ciertas variables de estado globales. Si es necesario leer
  parmetros o algo as, los lee en formato ASCII y los convierte a UTF-8.

  Return: (int) Devuelve cero con xito, o cualquier otro nmero cuando la
  funcin cree que el programa no merece continuar, por error grave.
  
***************************************************************************/

int      decodifica_linea_de_comandos (int argc, char *argv[])
{
   int f,ret;

   /* En el primer argumento est siempre el path. De l conseguimos la
    * ruta de ejecucin, y el nombre del ejecutable.
    */

   fix_filename_path (path_ejecucion, uconvert_ascii (argv[0], NULL), STR);
   replace_filename (directorio_programa, path_ejecucion, "", STR);
   append_filename (ejecutable_programa, "", get_filename (path_ejecucion), STR);
   replace_extension (fichero_records, path_ejecucion, "rec", STR);

   /* Ahora recorremos el resto de los parmetros, y vemos si nos interesan
    * ajustando bits globales en ese caso.
    */
   for (f=1; f<argc; f++)
   {
      if (ustricmp (argv[f], "-h")==0 || ustricmp (argv[f], "--help")==0)
         SET_FLAG (banderas, BIT_SHOW_HELP);
      else if (ustricmp (argv[f], "-r")==0 || ustricmp (argv[f], "--res")==0)
         SET_FLAG (banderas, BIT_CAMBIO_RES);
      else if (ustricmp (argv[f], "-l")==0 || ustricmp (argv[f], "--log")==0)
      {
         f++;
         if (f >= argc)
         {
            eprintf ("decodifica_linea_de_comandos: tienes que especificar un "
                     "fichero de logging");
            return 1;
         }

         ret = comprueba_validez_fichero_log (uconvert_ascii (argv[f], NULL) );
         if (ret)
            return ret;
      }
   }

   return 0;
}


/**[txh]********************************************************************

  Description: Funcion mnima que intenta entrar en un modo grfico seguro,
  para al menos poder seguir la ejecucin del programa. Entra en una
  resolucin 'no determinada' (cuidado con SCREEN_W y SCREEN_H), y siendo
  posiblemente el modo de 8 bits, ajusta la paleta. Tras esto, limpia la
  pantalla a color blanco, para que las letras se impriman en negro.
  Si el bit BIT_CAMBIO_RES est activo, muestra un men de seleccin para
  elegir una resolucin de pantalla.

  Return: (int) Cero con xito, cualquier otro nmero con fracaso. En este
  caso, el programa DEBE abortar, ya que es vital entrar en modo grfico.
  
***************************************************************************/


int      entra_en_modo_grafico (void)
{
   PALETTE pal;
   int ret, ancho, alto, card, w, h;

   /* Leemos del fichero de configuracin la resolucin sugerida */
   ancho = get_config_int (SECCION_ARKANOID, CFG_ANCHO_PANTALLA, 320);
   alto  = get_config_int (SECCION_ARKANOID, CFG_ALTO_PANTALLA, 200);

   /* Entramos en modo grfico */
   ret = set_gfx_mode (GFX_SAFE, ancho, alto, 0, 0);
   if (ret)
   {
      eprintf ("entra_en_modo_grafico: No se pudo establecer un modo grafico "
               "para el programa: %s", allegro_error);
      return 1;
   }

   /* Ponemos una paleta de colores */
   generate_332_palette (pal);
   /* Especificamos el color 0 como negro, manualmente */
   pal[0].r = pal[0].g = pal[0].b = 0;
   set_palette (pal);

   /* Limpiamos la pantalla de color blanco y activamos banderas */
   clear_to_color (screen, makecol (0, 0, 0));
   SET_FLAG (banderas, BIT_MODO_GRAFICO);

   if (GET_FLAG (banderas, BIT_CAMBIO_RES))
   {
      /* Mostramos dilogo de seleccin de resolucin */
      if (!gfx_mode_select (&card, &w, &h))
      {
         eprintf ("entra_en_modo_grafico: El usuario salio del programa.");
         return 2; /* Devolvemos error */
      }
      /* Tenemos los datos. Ahora probamos esa resolucin */
      ret = set_gfx_mode (card, w, h, 0, 0);
      if (ret)
      {
         eprintf ("entra_en_modo_grafico: no se pudo activar modo %dx%d",
                  w, h);
         return 3;
      }
      set_palette (pal);
      clear_to_color (screen, makecol (0, 0, 0));
   }

   /* Comprobamos que la pantalla no es ridculamente pequea */
   if (SCREEN_W < MINIMO_PANTALLA_X || SCREEN_H < MINIMO_PANTALLA_Y)
   {
      eprintf ("La resolucion seleccionada (%dx%d) es demasiado pequenia.",
               SCREEN_W, SCREEN_H);
      eprintf ("El minimo para jugar es %dx%d. Ejecuta el programa con -r.",
               MINIMO_PANTALLA_X, MINIMO_PANTALLA_Y);
      return 4;
   }

   ret = inicializa_motor_grafico();
   if (ret)
      return 5;

   text_mode (-1);   /* Textos transparentes */
   lprintf ("Modo grafico: %dx%d", SCREEN_W, SCREEN_H);
   return 0;
}


/**[txh]********************************************************************

  Description: Pequea funcin que hace de intermediaria con el hardware.
  Pone el modo texto (o grafico) anterior a la ejecucion del programa.

***************************************************************************/

void     sal_del_modo_grafico (void)
{
   if (GET_FLAG (banderas, BIT_MODO_GRAFICO))
   {
      /* Antes de salir, guardamos la resolucin actual en el fichero cfg */
      set_config_int (SECCION_ARKANOID, CFG_ANCHO_PANTALLA, SCREEN_W);
      set_config_int (SECCION_ARKANOID, CFG_ALTO_PANTALLA, SCREEN_H);
   
      CLEAR_FLAG (banderas, BIT_MODO_GRAFICO);
      if (set_gfx_mode (GFX_TEXT, 0, 0, 0, 0))
      {
         eprintf ("sal_del_modo_grafico: no se pudo volver al modo anterior (???)");
         eprintf (allegro_error);
      }
      finaliza_motor_grafico();
   }
}

/**[txh]********************************************************************

  Description: Especifica el fichero .cfg que se usar para salvar algunos
  datos del juego, como resolucin, fichero logging, etc. Lee automticamente
  del fichero algunos datos y los usa si es necesario, como por ejemplo el
  de un fichero logging especifico.

***************************************************************************/


void     lee_configuracion_del_fichero_cfg (void)
{
   char path_fichero_cfg[STR];
   char *fichero_log;

   /* Especificamos el fichero de configuracin */
   append_filename (path_fichero_cfg, directorio_programa, ARKANOID_CFG, STR);
   /* En caso de que el fichero no exista, forzamos la seleccin de resolucin */
   if (!exists(path_fichero_cfg))
      SET_FLAG (banderas, BIT_CAMBIO_RES);
   
   set_config_file (path_fichero_cfg);

   /* Comprobamos si hay que anular el logging */
   if (GET_FLAG (banderas, BIT_IGNORA_LOG))
   {
      set_config_string (SECCION_ARKANOID, CFG_FICH_LOG, "NULL");
      return ;
   }

   /* Miramos si se especific -l en la linea de comando */
   if (!GET_FLAG (banderas, BIT_LOG_A_FICHERO))
   {
      /* Vemos si se quiere hacer un logging */
      fichero_log = get_config_string (SECCION_ARKANOID, CFG_FICH_LOG, NULL);

      if (fichero_log) /* no nos interesa el valor de retorno */
         comprueba_validez_fichero_log (fichero_log);
   }
}



/**[txh]********************************************************************

  Description: Esta funcin es llamada desde @x{finaliza_programa} si el
  bit BIT_SHOW_HELP fue activada antes. Se supone que se est en modo texto
  y se quiere ver la ayuda rpida del programa. Mediante allegro_message
  se imprime lo pedido.

***************************************************************************/

void     muestra_ayuda_linea_de_comando (void)
{
   if (!GET_FLAG (banderas, BIT_SHOW_HELP))
      return ;

   imprime_informacion_sistema (0);
   
   allegro_message (
      "Parametros de ejecucion:\n"
      "-h, --help         Muestra version, ayuda y sale.\n"
      "-l xxx,            Almacena mensajes de logging en el fichero xxx. Si se usa\n"
      " --log xxx         NULL como nombre xxx, se desactivan los loggings.\n"
      "-r, --res          Solicita un cambio de resolucion.\n"
      );
}


