/*
 * Nombre del proyecto: Proyecto Arkanoid
 *
 * Nombre del fichero: juego.c
 *
 * Nombre de los autores: Grzegorz Adam Hankiewicz y Carlos Pereda Moreno
 *
 * Fecha: 18 de Enero del 2000
 *
 * Descripcin: Parte principal del cdigo que controla el juego y los mens.
 * Este cdigo es llamado por la funcin main, la cual delega toda
 * responsabilidad en la funcin juego.
 */

/* CABECERAS GLOBALES */
#ifdef MSS
#include <mss.h>
#endif MSS
#include <string.h>
#include <allegro.h>

/* CABECERAS DE FUNCIONES EXTERNAS */
#include "general.h"
#include "struct.h"
#include "errores.h"
#include "log.h"
#include "logico.h"
#include "graph.h"
#include "main.h"
#include "lborrado.h"
#include "hiscore.h"

/* CABECERA PROPIA */
#define DECLARA_VARIABLES_JUEGO_C
#include "juego.h"

int      opcion_menu = 0;

/**[txh]********************************************************************

  Description: Funcin que carga el mapa de ladrillos que sigue a
  continuacin, comprobando previamente que existe siguiente nivel, parando
  el juego mediante la activacin de los flags correspondientes si no existen
  ms mapas de ladrillos.

  Return: Devuelve 1 si el mapa de ladrillos se carg con xito; 0 en caso
  contrario.
  
***************************************************************************/

int      carga_nivel (DATOS_JUEGO *datos, int *nivel)
{
   BITMAP *fase;
   char temporal[STR];

   usprintf (temporal, "%s/levels/mapa%02d.pcx", directorio_programa, *nivel);

   /* Aumenta el nivel y activa la bandera de salida si se sobrepasa el lmite */
   (*nivel)++;
   if (*nivel > MAX_NIVELES)
      SET_FLAG (datos->parar_juego, BIT_SALIR_DE_JUEGO);
   
   /* Si fase = 0 => Error, xito en caso contrario */
   if (!(fase = carga_mapa_ladrillos (temporal)))
   {
      lprintf ("carga_nivel: error cargando mapa de ladrillos");
      return 0;
   }
   /* Destruimos el mapa antiguo */
   destroy_bitmap (datos->ladrillos);
   /* Cargamos el mapa nuevo */
   datos->ladrillos = fase;

   return 1;
}

/**[txh]********************************************************************

  Description: Inicializa los datos que tienen que permanecer constantes al
  principio de cada partida.

***************************************************************************/



void     inicializa_datos_fijos (DATOS_JUEGO *datos)
{
   datos->vidas = 3;
   datos->puntos = 0;
}

/**[txh]********************************************************************

  Description: Funcin que inicializa los datos y coordenadas al valor que
  deben tener al principio de cada nivel.

***************************************************************************/


void     inicializa_datos_nivel (DATOS_JUEGO *datos)
{
   /* Dibujamos los "bordes" de la pantalla */
   dibuja_marco_pantalla (screen);
   /* Establecemos las coordenadas iniciales de la pelota, la raqueta,
     sus incrementos correspondientes, e inicializamos variables */
   datos->parar_juego = 0;
   datos->x_p = itofix (RES_LOGICA_X/2);
   datos->y_p = F_ALTURA_RAQUETA - itofix (20);
   datos->x_r = itofix (RES_LOGICA_X/2);
   datos->angulo = itofix (32 + (rand()%64));
   datos->longitud = 0;

   marca_zona_borrado (lista_borrado_logica, 0, 0, RES_LOGICA_X, RES_LOGICA_Y);
   sec_counter = tick_counter = datos->tiempo = 0;
}

/**[txh]********************************************************************

  Description: Funcin  en la que se hacen las llamadas que inicializan los
  datos y se llama a la funcin "ncleo" que ejecuta el bucle principal del
  juego.

  Return: Devuelve 0 si todo se llev a cabo con xito. Distinto de 0 en caso
  contrario.

***************************************************************************/


int      bucle_juego (DATOS_JUEGO *datos)
{
   int nivel = 0;

   if (!datos)
   {
      eprintf ("bucle_juego: punteros nulos");
      return 1;
   }

   lprintf ("bucle_juego: inicializando datos partida");
   inicializa_datos_fijos (datos);

   while (!GET_FLAG (datos->parar_juego, BIT_GAME_OVER|BIT_SALIR_DE_JUEGO))
   {
      if (carga_nivel (datos, &nivel))
      {
         inicializa_datos_nivel (datos);
         procesa_juego (datos);
      }
      else
         lprintf ("El nivel %d no existe", nivel);
   }

   lprintf ("bucle_juego: se salio del bucle de juego. Puntos: %d",
      datos->puntos);
   /* Insertamos los puntos en los records si procede */
   score_table (datos->puntos);

   /* Limpiamos la variable que indica interrupcin del juego y la pantalla
      para restablecer el men */
   datos->parar_juego = 0;
   dibuja_menu ();
   clear_keybuf(); /* Ralentizacin artificial de teclado */

   return 0;
}

/**[txh]********************************************************************

  Description: Funcin "ncleo" que realiza el bucle del juego hasta que se
  activen o el BIT_GAME_OVER, o el BIT_SALIR_DE_JUEGO. Dentro del bucle se
  realizan todos los chequeos correspondientes a las reglas del juego, el
  bucle lgico y el redibujado de los elementos en pantalla.

***************************************************************************/


void     procesa_juego (DATOS_JUEGO *datos)
{
   int old_sec, cont_comenzar;

   old_sec = sec_counter;
   cont_comenzar = sec_counter;
   SET_FLAG (datos->parar_juego, BIT_ESPERAR_BOLA);
   
   /* Bucle que se ejecuta hasta que:
      - BIT_GAME_OVER = 1 => Fin de juego por acabarse todos las vidas;
      - BIT_SALIR_DE_JUEGO = 1 => Fin de juego por pulsar [ESC]; */
   while (!GET_FLAG (datos->parar_juego, BIT_GAME_OVER|BIT_SALIR_DE_JUEGO))
   {
      /* Detectamos teclas que pulsa el usuario hasta que tick_counter = 0 */
      do
      {
         recoge_entrada_usuario (datos);
      } while (!tick_counter);

      /* Bucle que espera a poner la bola en movimiento, antes de empezar */
      if (GET_FLAG (datos->parar_juego, BIT_ESPERAR_BOLA))
      {
         if (sec_counter > cont_comenzar+1)
         {
            CLEAR_FLAG (datos->parar_juego, BIT_ESPERAR_BOLA);
            datos->longitud = itofix (1);
         }
      }

      datos->tiempo = tick_counter;
      tick_counter = 0;
      /* Ejecutamos el bucle lgico hasta que tiempo = 0 */
      while (datos->tiempo > 0)
      {
         bucle_logico (datos);
         datos->tiempo--;
      }
      
      /* Se acab el nivel */
      if (GET_FLAG (datos->parar_juego, BIT_SIGUIENTE_NIVEL))
         return ;

      /* Si se activa BIT_PERDER_VIDA, las vidas se decrementan, y se
         inicializan los datos correspondientes para volver a empezar
         en el nivel. Si vidas = 0 => BIT_GAME_OVER => fin de juego */
      if (GET_FLAG (datos->parar_juego, BIT_PERDER_VIDA))
      {
         datos->vidas--;
         inicializa_datos_nivel (datos);
         if (datos->vidas < 1)
            SET_FLAG (datos->parar_juego, BIT_GAME_OVER);

         SET_FLAG (datos->parar_juego, BIT_ESPERAR_BOLA);
         cont_comenzar = sec_counter;
         datos->longitud = 0;
      }
      /* Dibujamos todos los bitmaps en la pantalla */
      dibuja_pantalla (datos);
   }
}

/**[txh]********************************************************************

  Description: Funcin principal del juego. Se diferencia de main por
  varias razones, siendo la ms importante que el ordenador ya est
  "aislado" o "abstrado" por Allegro y dems funciones. Es decir, aqu ya
  estamos en un modo grfico til, con variables inicializadas, no hace
  falta llamar finaliza_programa desde ninguna parte (se hace desde fuera),
  etc. Adems, esto controla principalmente los mens y el bucle del juego.

  Return: (int) Cero si el juego se ejecut con xito, otro nmero si hubo
  problemas.
  
***************************************************************************/

int      juego (void)
{
   int ret;
   DATOS_JUEGO *datos;

   lprintf ("Creando estructura de datos jugador");
   datos = inicializa_datos_juego();
   if (!datos)
      return 2;

   dibuja_menu ();
   
   /* Bucle de seleccin de opciones en el men */
   while (!datos->parar_juego)
   {
      recoge_entrada_usuario (datos);

      ret = opcion_menu;
      
      if (datos->mov_y < 0) opcion_menu --;
      if (datos->mov_y > 0) opcion_menu ++;
      if (opcion_menu < 0) opcion_menu = 2;
      if (opcion_menu > 2) opcion_menu = 0;
      
      if (ret != opcion_menu)
      {
         vsync();
         dibuja_menu ();
         clear_keybuf(); /* Ralentizacin artificial de respuesta del teclado */
      }
         
      /* Se ha seleccionado la opcin? */
      if (GET_FLAG (datos->botones, BT_DISP1))
      {
         switch (opcion_menu)
         {
            case 0:
               ret = bucle_juego (datos);
               if (ret)
                  return ret;
               break;
            case 1:
               score_table (0);
               dibuja_menu ();
               clear_keybuf(); /* Ralentizacin artificial del teclado */
               break;
            case 2:
               datos->parar_juego++;
               break;
            default:
               eprintf ("juego: se llego a opcion default");
               break;
         }
      }
   }

   lprintf ("Desinicializando juego...");
   /* Limpiamos todas las variables antes de salir del juego */
   destruye_datos_juego (datos);

   return 0;
}


/**[txh]********************************************************************

  Description: Funcin que crea una estructura DATOS_JUEGO y la pone a
  cero.

  Return: (DATOS_JUEGO*) Puntero a estructura. NULL si hubo problemas.

***************************************************************************/

DATOS_JUEGO *inicializa_datos_juego (void)
{
   DATOS_JUEGO *datos;
   int i;

   /* Reservamos espacio para la estructura DATOS_JUEGO y comprobamos que
      exista espacio para esta */
   datos = malloc (sizeof (DATOS_JUEGO));
   if (!datos)
   {
      eprintf ("inicializa_datos_juego: no hay memoria libre");
      return NULL;
   }

   /* Inicializamos todos los datos a cero */
   memset (datos, 0, sizeof (DATOS_JUEGO));

   /* Creamos los bitmaps de ladrillos, bmp, bola y raqueta, comprobamos que
      hay espacio para estos y los inicializamos todos */
   datos->ladrillos = create_bitmap (RES_LOGICA_X/TAM_CELDA,
                                     RES_LOGICA_Y/TAM_CELDA);
   if (!datos->ladrillos)
   {
      destruye_datos_juego (datos);
      eprintf ("inicializa_datos_juego: no hay memoria libre");
      return NULL;
   }
   clear (datos->ladrillos);

   datos->bmp = create_bitmap (RES_LOGICA_X, RES_LOGICA_Y);
   if (!datos->bmp)
   {
      destruye_datos_juego (datos);
      eprintf ("inicializa_datos_juego: no hay memoria libre");
      return NULL;
   }
   
   datos->bola = create_bitmap (ANCHO_BOLA, ALTO_BOLA);
   if (!datos->bola)
   {
      destruye_datos_juego (datos);
      eprintf ("inicializa_datos_juego: no hay memoria libre");
      return NULL;
   }

   datos->raqueta = create_bitmap (ANCHO_RAQUETA, ALTO_RAQUETA);
   if (!datos->raqueta)
   {
      destruye_datos_juego (datos);
      eprintf ("inicializa_datos_juego: no hay memoria libre");
      return NULL;
   }

   clear_to_color (datos->bmp, makecol (0, 0, 0));
   clear (datos->bola);
   clear (datos->raqueta);
   
   circlefill (datos->bola,ANCHO_BOLA/2, ALTO_BOLA/2, ANCHO_BOLA/2,
               makecol (0, 250, 0));
   for (i = ALTO_RAQUETA/2; i < ANCHO_RAQUETA - ALTO_RAQUETA/2; i++)
      circlefill (datos->raqueta, i, ALTO_RAQUETA/2,
                  ALTO_RAQUETA/2, makecol (250, 0, 0));

   return datos;
}


/**[txh]********************************************************************

  Description: Funcin que destruye una estructura de DATOS_JUEGO antes
  creada por @x{inicializa_datos_juego}.

***************************************************************************/


void     destruye_datos_juego (DATOS_JUEGO *datos)
{
   if (!datos)
      return;

   if (datos->ladrillos)
      destroy_bitmap (datos->ladrillos);

   if (datos->bmp)
      destroy_bitmap (datos->bmp);
      
   if (datos->bola)
      destroy_bitmap (datos->bola);
      
   if (datos->raqueta)
      destroy_bitmap (datos->raqueta);
      
   free (datos);
}



/**[txh]********************************************************************

  Description: Carga en memoria un fichero grfico que contiene el mapa de
  ladrillos. Pasar como parmetro el nombre del fichero grfico a cargar.
  El fichero grfico debe estar a una profundidad de color de 8 bits (256
  colores), y debe tener un mnimo de tamao de RES_LOGICA_X / TAM_CELDA
  por RES_LOGICA_Y / TAM_CELDA (que actualmente equivale a una imgen de
  52x40). En otras palabras: si la imgen es ms grande, slo se usar la
  esquina superior izquierda, el resto ser descartado.

  Return: (BITMAP *) Devuelve puntero al mapa de ladrillos o NULL si hubo
  fallo.

***************************************************************************/


BITMAP   *carga_mapa_ladrillos (char *texto)
{
   PALETTE pal;
   BITMAP *carga, *temporal;
   
   if (!texto)
   {
      eprintf ("carga_mapa_ladrillos: Puntero nulo");
      return NULL;
   }

   lprintf ("carga_mapa_ladrillos: cargando '%s'", texto);
   /* Comprobamos si el fichero existe */
   if (!file_exists (texto, 0, NULL))
   {
      lprintf ("carga_mapa_ladrillos: el fichero '%s' no existe", texto);
      return NULL;
   }

   /* Lo cargamos */
   carga = load_bitmap (texto, pal);
   if (!carga)
   {
      eprintf ("carga_mapa_ladrillos: load_bitmap devolvi NULL");
      return NULL;
   }

   /* Comprobamos que es un bitmap vlido para nuestro uso */
   if (bitmap_color_depth (carga) != 8)
   {
      destroy_bitmap (carga);
      lprintf ("carga_mapa_ladrillos: el fichero '%s' no es de 8 bits", texto);
      return NULL;
   }

   if (carga->w < (RES_LOGICA_X / TAM_CELDA) || carga->h < (RES_LOGICA_Y / TAM_CELDA))
   {
      lprintf ("carga_mapa_ladrillos: el fichero '%s' tiene un tamanio de %dx%d, "
               "cuando deberia ser de al menos %dx%d", texto, carga->w, carga->h,
               (RES_LOGICA_X / TAM_CELDA),(RES_LOGICA_Y / TAM_CELDA));
      destroy_bitmap (carga);
      return NULL;
   }

   if (carga->w != (RES_LOGICA_X / TAM_CELDA) || carga->h != (RES_LOGICA_Y / TAM_CELDA))
   {
      lprintf ("carga_mapa_ladrillos: el fichero '%s' es demasiado grande: "
         "%dx%d, recortando superficie...", texto, carga->w, carga->h);

      temporal = create_bitmap (RES_LOGICA_X / TAM_CELDA, RES_LOGICA_Y / TAM_CELDA);
      if (!temporal)
      {
         destroy_bitmap (carga);
         eprintf ("carga_mapa_ladrillos: insuficiente memoria");
         return NULL;
      }

      blit (carga, temporal, 0, 0, 0, 0, temporal->w, temporal->h);
      destroy_bitmap (carga);
      carga = temporal;
   }

   lprintf ("carga_mapa_ladrillos: mapa obtenido correctamente.");
   return carga;
}

/**[txh]********************************************************************

  Description: Funcin que muestra el men con las opciones en pantalla.

***************************************************************************/

void     dibuja_menu (void)
{
   BITMAP *bmp;
   int h, col;

   bmp = create_bitmap (SCREEN_W, SCREEN_H);
   if (!bmp)
   {
      eprintf ("dibuja_menu: no hay suficiente memoria.");
      return ;
   }
   clear (bmp);

   /* Ahora lo dibujamos todo en el bitmap temporal */
   h = SCREEN_H / 7;

   if (opcion_menu == 0)
      col = makecol (0, 128, 255);
   else
      col = makecol (0, 255, 0);
   escribe_texto (bmp, font, 0, h, SCREEN_W, h, col, 0, "   JUGAR   ");
   
   if (opcion_menu == 1)
      col = makecol (0, 128, 255);
   else
      col = makecol (0, 255, 0);
   escribe_texto (bmp, font, 0, 3*h, SCREEN_W, h, col, 0, "  RECORDS  ");
   
   if (opcion_menu == 2)
      col = makecol (0, 128, 255);
   else
      col = makecol (0, 255, 0);
   escribe_texto (bmp, font, 0, 5*h, SCREEN_W, h, col, 0, "   SALIR   ");

   /* Finalmente lo mostramos y destrumos el bitmap temporal */
   blit (bmp, screen, 0, 0, 0, 0, bmp->w, bmp->h);
   destroy_bitmap (bmp);
}

